/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	parser													 */
/* --------------------------------------------------------- */

#include "parser.h"
#include "variable_manager.h"
#include "system_resource.h"
#include "zb_error.h"
#include "debug.h"
#include <cstring>
#include <cctype>

using namespace std;

/* --------------------------------------------------------- */
namespace {
	unsigned int analyze_line_no( const char* &p_string ) {
		unsigned int num = 0, columns = 0;

		while( isdigit( *p_string ) && (columns < 5) ) {
			num = num * 10 + (*p_string - '0');
			p_string++;
			columns++;
		}
		if( isdigit( *p_string ) ) return 0;
		return num;
	}
}

/* --------------------------------------------------------- */
c_parser::c_parser( void ) {

	length = 0;
	line_no = 0;
	std::memset( precompiled_code, 0, sizeof(precompiled_code) );
	std::memset( line, 0, sizeof(line) );
}

/* --------------------------------------------------------- */
void c_parser::analyze_number( int& index, const char* &p_string ) {
	int column;
	unsigned long i_number = 0;
	double r_number = 0., weight;

	if( p_string[0] == '0' ) {
		if( p_string[1] == 'x' ) {
			/* hexadecimal number */
			p_string += 2;
			while( isxdigit( p_string[0]  ) ) {
				if( isdigit( p_string[0]  ) ) {
					i_number = (i_number << 4) | (p_string[0]  - '0');
				}
				else {
					i_number = (i_number << 4) | (toupper(p_string[0] ) - 'A' + 10);
				}
				p_string++;
			}
			precompiled_code[ index++ ] = PC_INUMBER;
			memcpy( precompiled_code + index, &i_number, sizeof(i_number) );
			index += sizeof(i_number);
			dlog.printf( "    Detect: hex number (0x%08X)\n", i_number );
			return;
		}
		else if( p_string[1] == 'b' ) {
			/* binary number */
			p_string += 2;
			while( (*p_string == '0') | (p_string[0]  == '1') ) {
				i_number = (i_number << 1) | (p_string[0]  - '0');
				p_string++;
			}
			precompiled_code[ index++ ] = PC_INUMBER;
			memcpy( precompiled_code + index, &i_number, sizeof(i_number) );
			index += sizeof(i_number);
			dlog.printf( "    Detect: binary number (0x%08X)\n", i_number );
			return;
		}
	}

	for( column = 0; isdigit( p_string[ column ] ); column++ ) {
	}
	if( p_string[ column ] == '.' ) {
		/* real number */
		while( isdigit( p_string[0]  ) ) {
			r_number = (r_number * 10) + (p_string[0] - '0');
			p_string++;
		}
		p_string++;
		weight = 0.1;
		while( isdigit( p_string[0]  ) ) {
			r_number = r_number + weight * (p_string[0] - '0');
			weight = weight / 10;
			p_string++;
		}
		precompiled_code[ index++ ] = PC_RNUMBER;
		memcpy( precompiled_code + index, &r_number, sizeof(r_number) );
		index += sizeof(r_number);
		dlog.printf( "    Detect: real number (%f)\n", r_number );
	}
	else {
		/* integer number */
		while( isdigit( p_string[0]  ) ) {
			i_number = (i_number * 10) + (p_string[0] - '0');
			p_string++;
		}
		precompiled_code[ index++ ] = PC_INUMBER;
		memcpy( precompiled_code + index, &i_number, sizeof(i_number) );
		index += sizeof(i_number);
		dlog.printf( "    Detect: decimal number (%d)\n", i_number );
	}
}

/* --------------------------------------------------------- */
int c_parser::get_operator_id( const char * &p_operator ) {
	unsigned int i, j;
	const char *p_operator_list[] = {
		"AND",	"OR",	"XOR",	"NOT",	"MOD",
		"+",	"-",	"*",	"/",	"(",
		")",	"=",	"<=",	">=",	"<>",
		"<<",	">>",	"<",	">",	":",
		";",	",",	"^",
	};

	for( i = 0; i < sizeof(p_operator_list)/sizeof(p_operator_list[0]); i++ ) {
		for( j = 0; p_operator_list[i][j] && p_operator[j]; j++ ) {
			if( toupper( p_operator[j] ) != p_operator_list[i][j] ) {
				break;
			}
		}
		if( p_operator_list[i][j] == 0 ) {
			p_operator += j;
			dlog.printf( "    Detect: operator(%s)\n", p_operator_list[i] );
			return i + PC_OPERATOR;	/* convert to OP_XXXX and return. */
		}
	}
	return -1;
}

/* --------------------------------------------------------- */
void c_parser::analyze_symbol( int& index, const char* &p_string ) {
	const char *p_reserve_word_list[] = {
		"for", "to", "step", "next", "if", "then", "else", "goto", "gosub", "return"
	};
	size_t length;
	int id;

	for( id = 0; id < static_cast<int>(sizeof(p_reserve_word_list)/sizeof(p_reserve_word_list[0]));
			id++ ) {
		length = strlen( p_reserve_word_list[id] );
		if( (strncmp( p_string, p_reserve_word_list[id], length ) == 0) && (! isalpha(p_string[length])) ) {
			precompiled_code[ index++ ] = static_cast<unsigned char>( id + PC_FOR );
			p_string += length;
			dlog.printf( "    Detect: command \"%s\"\n", p_reserve_word_list[ id ] );
			return;
		}
	}

	c_variable *p;
	id = get_operator_id( p_string );
	if( id != -1 ) {
		/* it is operator. */
		precompiled_code[ index++ ] = static_cast<unsigned char>( id );
		return;
	}
	/* it is valiable, command or function. */
	char word[ MAX_VARIABLE_LENGTH + 1 ];
	int i = 0;

	do {
		word[ i ] = p_string[ i ];
		i++;
		if( i >= MAX_VARIABLE_LENGTH ) break;
	} while( is_name_char(p_string[ i ]) );
	word[ i ] = 0;

	while( isspace( p_string[ i ] ) ) {
		i++;
	}

	id = p_res->p_interpreter->get_command_code( word );
	if( id != -1 ) {
		/* command */
		precompiled_code[ index++ ] = PC_COMMAND;
		memcpy( &precompiled_code[ index ], &id, sizeof(int) );
		index += sizeof(int);
		p_string += i;
		return;
	}

	id = p_res->p_interpreter->get_function_code( word );
	if( id != -1 ) {
		/* function */
		precompiled_code[ index++ ] = PC_FUNCTION;
		memcpy( &precompiled_code[ index ], &id, sizeof(int) );
		index += sizeof(int);
		p_string += i;
		return;
	}

	if( p_string[ i ] == '(' ) {
		/* array */
		switch( toupper( word[0] ) ) {
		case 'I':
			p = p_res->p_variable_manager->add_array< c_variable_integer_array >( word );
			precompiled_code[ index++ ] = PC_IARRAY;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(integer) \"%s\"\n", word );
			break;
		case 'R':
			p = p_res->p_variable_manager->add_array< c_variable_real_array >( word );
			precompiled_code[ index++ ] = PC_IARRAY;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(real) \"%s\"\n", word );
			break;
		case 'S':
			p = p_res->p_variable_manager->add_array< c_variable_string_array >( word );
			precompiled_code[ index++ ] = PC_IARRAY;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(string) \"%s\"\n", word );
			break;
		default:
			p_res->error_code = c_zb_error::ERROR_SYNTAX;
			break;
		}
	}
	else {
		/* variable */
		switch( toupper( word[0] ) ) {
		case 'I':
			p = p_res->p_variable_manager->add_single< c_variable_integer >( word );
			precompiled_code[ index++ ] = PC_IVAR;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(integer array) \"%s\"\n", word );
			break;
		case 'R':
			p = p_res->p_variable_manager->add_single< c_variable_real >( word );
			precompiled_code[ index++ ] = PC_RVAR;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(real array) \"%s\"\n", word );
			break;
		case 'S':
			p = p_res->p_variable_manager->add_single< c_variable_string >( word );
			precompiled_code[ index++ ] = PC_SVAR;
			index += c_parser::put_ptr( &precompiled_code[ index ], p );
			dlog.printf( "    Detect: variable(string array) \"%s\"\n", word );
			break;
		default:
			p_res->error_code = c_zb_error::ERROR_SYNTAX;
			break;
		}
	}
	p_string += i;
}

/* --------------------------------------------------------- */
void c_parser::analyze_operator( int& index, const char* &p_string ) {

	int id = get_operator_id( p_string );
	if( id == -1 ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	precompiled_code[ index++ ] = static_cast<unsigned char>( id );
}

/* --------------------------------------------------------- */
void c_parser::analyze_string( int& index, const char* &p_string ) {
	char line[ 255 ];
	int length = 0;

	p_string++;
	while( *p_string != 0 && (length < 255) ) {
		if( *p_string == '"' ) {
			if( p_string[1] == '"' ) {
				line[ length ] = '"';
				length++;
				p_string += 2;
			}
			else {
				break;
			}
		}
		else {
			line[ length ] = *p_string;
			length++;
			p_string++;
		}
	}

	if( *p_string != '"' ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_string++;
	precompiled_code[ index++ ] = PC_STRING;
	precompiled_code[ index++ ] = static_cast<unsigned char>( length );
	if( length ) {
		memcpy( precompiled_code + index, line, length );
	}
	index += length;
}

/* --------------------------------------------------------- */
void c_parser::analyze_label( int& index, const char* &p_string, bool f_set, c_program_line* p_line ) {
	c_variable_label *p;
	char word[ MAX_VARIABLE_LENGTH + 1 ];
	int i = 0;

	p_string++;
	do {
		word[ i ] = p_string[ i ];
		i++;
	} while( is_name_char(p_string[ i ]) );
	word[ i ] = 0;

	p = p_res->p_variable_manager->add_label( word );
	dlog.printf( "    Detect: label %s (%p)\n", word, p );
	precompiled_code[ index++ ] = PC_LABEL;
	index += c_parser::put_ptr( &precompiled_code[ index ], p );
	p_string += i;

	if( f_set ) {
		dlog.printf( "      Instance: set label instance\n" );
		p->set( index );
		p->set_line( p_line );
	}
}

/* --------------------------------------------------------- */
int c_parser::check_line_no( const char* &p_string ) {

	p_res->error_code = c_zb_error::NO_PROBLEM;

	/* skip white space */
	while( isspace( *p_string ) ) {
		p_string++;
	}

	if( *p_string == 0 ) return -1;		/* no operation */
	if( isdigit(*p_string) ) {
		line_no = analyze_line_no( p_string );
		if( line_no == 0 ) {
			p_res->error_code = c_zb_error::ERROR_ILLEGAL_LINE_NO;
			return 0;
		}
	}
	return line_no;		/* 0 is direct run. */
}

/* --------------------------------------------------------- */
int c_parser::precompile( const char *p_string, c_program_line* p_line ) {
	int index = 0;
	bool f_set = true;
	const char *p;

	dlog.printf( "Start precompile.\n" );
	p = p_string;
	while( isspace(*p) ) {
		p++;
	}
	if( *p == '\0' ) {
		/* blank */
		dlog.printf( "End precompile (blank).\n" );
		return -1;
	}

	dlog.printf( "  Line: %s\n", p_string );
	p_res->error_code = c_zb_error::NO_PROBLEM;

	strcpy( line, p_string );
	while( *p_string && (p_res->error_code == c_zb_error::NO_PROBLEM) ) {
		if( isdigit( *p_string ) ) {
			f_set = false;
			analyze_number( index, p_string );
		}
		else if( isalpha( *p_string ) ) {
			f_set = false;
			analyze_symbol( index, p_string );
		}
		else if( *p_string == '"' ) {
			f_set = false;
			analyze_string( index, p_string );
		}
		else if( *p_string == '$' ) {
			analyze_label( index, p_string, f_set, p_line );
		}
		else if( *p_string == '\'' ) {
			dlog.printf( "    Detect: comment\n" );
			break;
		}
		else if( ! isspace( *p_string ) ) {
			f_set = ( *p_string == ':' );
			analyze_operator( index, p_string );
		}
		else {
			/* white space */
			p_string++;
		}
	}
	dlog.printf( "    Endmark\n" );
	precompiled_code[ index++ ] = PC_ENDMARK;

	if( p_res->error_code != c_zb_error::NO_PROBLEM ) {
		p_res->p_variable_manager->line_release( precompiled_code, index );
		precompiled_code[ 0 ] = PC_ENDMARK;
		length = 0;
		return 0;
	}

	length = index;
	dlog.printf( "End precompile(line %d)\n", line_no );
	return line_no;		/* 0 is direct run. */
}

/* --------------------------------------------------------- */
void c_parser::variable_release( void ) {

	p_res->p_variable_manager->line_release( precompiled_code, length );
}
